import io
import os
import re
import shutil
import string
import sys
import tempfile
import pickle
import subprocess
from itertools import product
from contextlib import nullcontext
import importlib
import uuid
import copy
from functools import partial

import pandas as pd
import numpy as np
import pytest
import joblib

import psweep as ps

pj = os.path.join
here = os.path.abspath(os.path.dirname(__file__))


# ----------------------------------------------------------------------------
# helpers
# ----------------------------------------------------------------------------


def system(cmd):
    return ps.system(cmd).stdout.decode()


def func_a(pset):
    # We need to multiply by a float here to make sure the 'result' column has
    # float dtype. Else the column will be cast to float once we add NaNs,
    # which breaks df.equals(other_df) .
    return {"result": pset["a"] * 10.0}


def find_examples(skip=["batch_dask"]):
    found_paths = []
    dr = os.path.abspath(f"{here}/../examples")
    for basename in os.listdir(dr):
        do_skip = False
        path = pj(dr, basename)
        for pattern in skip:
            if pattern in basename:
                print(f"skipping {path} because of skip {pattern=}")
                do_skip = True
                break
        if not do_skip:
            found_paths.append(path)
    return found_paths


# ----------------------------------------------------------------------------
# test function
# ----------------------------------------------------------------------------


@pytest.mark.parametrize("path", find_examples())
def test_run_all_examples(path):
    # Running examples via shell will swallow all warnings, but hey, one
    # problem at a time.
    print(f"running example: {path}")
    if path.endswith(".py"):
        with tempfile.TemporaryDirectory() as tmpdir:
            cmd = f"""
                cp {path} {tmpdir}/ && cd {tmpdir} && \
                python3 {path}
            """
            print(system(cmd))
    elif os.path.isdir(path):
        with tempfile.TemporaryDirectory() as tmpdir:
            shutil.copytree(path, tmpdir, dirs_exist_ok=True)
            cmd = f"cd {tmpdir} &&./run_example.sh &&./clean.sh"
            print(system(cmd))


def test_shell_call_fail():
    """
    When calling multiple shell commands, we need to chain them using
      cmd1 && cmd2 && ... && cmdN
    rather than
      cmd1; cmd2; ...; cmdN
    In the latter case, only cmdN's exit code determines the returncode, thus
    hiding any previously failed commands.
    """
    # this must pass
    system("ls; thiswillfail but be hidden; ls; pwd")

    # all others must fail
    with pytest.raises(subprocess.CalledProcessError):
        system("ls && thiswillfail and break the chain && ls && pwd")

    # several types of script content that must fail
    txts = []
    txts.append(
        """
echo "exit 1"
exit 1
"""
    )

    txts.append(
        """
set -eux
exit 1111
"""
    )

    txts.append(
        """
set -eux
randomfoocommanddoesntexist
"""
    )

    txts.append(
        """
set -eux
python3 -c "raise Exception('foo')"
"""
    )

    basename = "test.sh"
    for txt in txts:
        with tempfile.TemporaryDirectory() as tmpdir:
            with pytest.raises(subprocess.CalledProcessError):
                ps.file_write(pj(tmpdir, basename), txt)
                print(system(f"cd {tmpdir}; sh {basename}"))


def test_shell_call():
    print(system("ls"))
    with tempfile.TemporaryDirectory() as tmpdir:
        txt = "ls"
        basename = "test.sh"
        ps.file_write(pj(tmpdir, basename), txt)
        print(system(f"cd {tmpdir}; sh {basename}"))


def test_run():
    with tempfile.TemporaryDirectory() as tmpdir:
        params = [{"a": 1}, {"a": 2}, {"a": 3}, {"a": 4}]
        calc_dir = f"{tmpdir}/calc"

        # run two times, updating the database, the second time,
        # also write tmp results
        df = ps.run(func_a, params, calc_dir=calc_dir)
        assert len(df) == 4
        assert len(df._run_id.unique()) == 1
        assert len(df._pset_id) == 4
        assert len(df._pset_id.unique()) == 4
        assert len(df._pset_hash) == 4
        assert len(df._pset_hash.unique()) == 4
        # For serial runs, this is guaranteed. For parallel runs, probably
        # also, as long as we use only concurrent.futures style APIs, avoid
        # map_async() and get results back in order. Should be safe with
        # multiprocessing, not sure about dask, though.
        assert (df._pset_seq.values == df.index.values).all()

        df = ps.run(
            func_a, params, calc_dir=calc_dir, poolsize=2, tmpsave=True
        )
        assert len(df) == 8
        assert len(df._run_id.unique()) == 2
        assert len(df._pset_id) == 8
        assert len(df._pset_id.unique()) == 8
        assert len(df._pset_hash) == 8
        assert len(df._pset_hash.unique()) == 4
        assert (df._pset_seq.values == df.index.values).all()
        assert set(df.columns) == set(
            [
                "_calc_dir",
                "_pset_id",
                "_run_id",
                "_pset_seq",
                "_run_seq",
                "_pset_hash",
                "_time_utc",
                "_pset_runtime",
                "_exec_host",
                "a",
                "result",
            ]
        )

        dbfn = f"{calc_dir}/database.pk"
        assert os.path.exists(dbfn)
        assert df.equals(ps.df_read(dbfn))

        # tmp results of second run
        run_id = df._run_id.unique()[-1]
        df_sel = df[df._run_id == run_id]
        for pset_id in df_sel._pset_id.values:
            tmpsave_fn = f"{calc_dir}/tmpsave/{run_id}/{pset_id}.pk"
            assert os.path.exists(tmpsave_fn)
            read_dct = ps.pickle_read(tmpsave_fn)
            assert isinstance(read_dct, dict)
            assert (
                df_sel[df_sel._pset_id == pset_id].iloc[0].to_dict()
                == read_dct
            )


@pytest.mark.parametrize("use_disk", [True, False])
def test_run_skip_dups(use_disk):
    params = [{"a": 1}, {"a": 2}, {"a": 3}, {"a": 4}]
    Ctx = tempfile.TemporaryDirectory if use_disk else nullcontext
    with Ctx() as tmpdir:
        # calc_dir is passed but not used if save=use_disk=False, so it
        # doesn't matter that in this case tmpdir=None and so
        # calc_dir="None/calc" as long as it is a string. The dir doesn't have
        # to exist.
        calc_dir = f"{tmpdir}/calc"

        df1 = ps.run(
            func_a,
            params,
            calc_dir=calc_dir,
            save=use_disk,
        )
        assert len(df1) == 4
        assert len(df1._run_id.unique()) == 1
        assert len(df1._pset_id.unique()) == 4

        # Run again w/ same params, but now skip_dups=True. This
        # will cause no psets to be run. That's why df2 is equal to df1.
        df2 = ps.run(
            func_a,
            params,
            calc_dir=calc_dir,
            df=None if use_disk else df1,
            save=use_disk,
            skip_dups=True,
        )
        assert df2.equals(df1)

        # Now use params where a subset is new (last 2 entries).
        params = [{"a": 1}, {"a": 2}, {"a": 88}, {"a": 99}]
        df3 = ps.run(
            func_a,
            params,
            calc_dir=calc_dir,
            df=None if use_disk else df2,
            save=use_disk,
            skip_dups=True,
        )
        assert len(df3) == 6
        assert len(df3._run_id.unique()) == 2
        assert len(df3._pset_id.unique()) == 6
        assert (
            df3._pset_hash.to_list()
            == df1._pset_hash.to_list() + df3._pset_hash.to_list()[-2:]
        )


def test_simulate():
    with tempfile.TemporaryDirectory() as tmpdir:
        params = [{"a": 1}, {"a": 2}, {"a": 3}, {"a": 4}]
        params_sim = [{"a": 88}, {"a": 99}]
        calc_dir = f"{tmpdir}/calc"
        calc_dir_sim = calc_dir + ".simulate"

        df = ps.run(func_a, params, calc_dir=calc_dir)
        df_sim = ps.run(func_a, params_sim, calc_dir=calc_dir, simulate=True)
        dbfn = f"{calc_dir}/database.pk"
        dbfn_sim = f"{calc_dir_sim}/database.pk"

        assert len(df_sim) == 6
        assert len(df) == 4
        assert os.path.exists(dbfn)
        assert os.path.exists(dbfn_sim)
        assert df.equals(ps.df_read(dbfn))
        assert df_sim.equals(ps.df_read(dbfn_sim))

        assert df.iloc[:4].equals(df_sim.iloc[:4])
        assert np.isnan(df_sim.result.values[-2:]).all()

        df2 = ps.run(func_a, params_sim, calc_dir=calc_dir)
        assert len(df2) == 6
        assert df.iloc[:4].equals(df2.iloc[:4])
        assert (df2.result.values[-2:] == np.array([880.0, 990.0])).all()


def test_is_seq():
    no = [{"a": 1}, io.IOBase(), "123"]
    yes = [[1, 2], {1, 2}, (1, 2)]
    for obj in no:
        print(obj)
        assert not ps.is_seq(obj)
    for obj in yes:
        print(obj)
        assert ps.is_seq(obj)


def test_df_io():
    from pandas.testing import assert_frame_equal

    letters = string.ascii_letters
    ri = np.random.randint
    rn = np.random.rand

    # random string
    def rs(n):
        return "".join(letters[ii] for ii in ri(0, len(letters), n))

    for fmt in ["pickle", "json"]:
        df = pd.DataFrame()
        for _ in range(2):
            vals = [
                ri(0, 100),
                rs(5),
                np.nan,
                f'"{rs(5)}"',
                f"'{rs(5)}'",
                (ri(0, 99), rn(), f"{rs(5)}"),
                [ri(0, 99), rn(), f"{rs(5)}"],
                rn(),
                rn(5),
                rn(5, 5),
                list(rn(5)),
                {"a": 1, "b": 3, "c": [1, 2, 3]},
            ]
            if fmt == "pickle":
                vals += [
                    True,
                    False,
                    None,
                    set(ri(0, 99, 10)),
                ]
            row = pd.DataFrame([dict(zip(letters, vals))])
            df = pd.concat((df, row), ignore_index=True)

        if fmt == "json":
            for orient in [
                None,
                "split",
                "records",
                "index",
                "columns",
                "_default_",
            ]:
                print("orient: ", orient)
                with tempfile.NamedTemporaryFile() as fd:
                    if orient != "_default_":
                        ps.df_write(fd.name, df, fmt=fmt, orient=orient)
                        read = ps.df_read(fd.name, fmt=fmt, orient=orient)
                    else:
                        ps.df_write(fd.name, df, fmt=fmt)
                        read = ps.df_read(fd.name, fmt=fmt)
                    assert_frame_equal(df, read, check_exact=False)
        elif fmt == "pickle":
            with tempfile.NamedTemporaryFile() as fd:
                ps.df_write(fd.name, df, fmt=fmt)
                read = ps.df_read(fd.name, fmt=fmt)
                assert_frame_equal(df, read)
        else:
            raise Exception("unknown fmt")


def test_save():
    with tempfile.TemporaryDirectory() as tmpdir:
        params = [{"a": 1}, {"a": 2}, {"a": 3}, {"a": 4}]
        calc_dir = f"{tmpdir}/calc"
        dbfn = f"{calc_dir}/database.pk"

        ps.run(func_a, params, calc_dir=calc_dir, save=False)
        assert not os.path.exists(dbfn)
        assert os.listdir(tmpdir) == []

        ps.run(func_a, params, calc_dir=calc_dir, save=True)
        assert os.path.exists(dbfn)
        assert os.listdir(tmpdir) != []


def test_merge_dicts():
    a = {"a": 1}
    b = {"b": 2}
    c = {"c": 3}

    # API
    assert a == ps.merge_dicts(a)
    m1 = ps.merge_dicts(a, b, c)
    m2 = ps.merge_dicts([a, b, c])

    # correct merge
    assert set(m1.keys()) == set(m2.keys())
    assert set(m1.values()) == set(m2.values())
    assert set(m1.keys()) == set(("a", "b", "c"))
    assert set(m1.values()) == set((1, 2, 3))

    # left-to-right
    a = {"a": 1}
    b = {"a": 2}
    c = {"a": 3}
    m = ps.merge_dicts(a, b)
    assert m == {"a": 2}
    m = ps.merge_dicts(a, b, c)
    assert m == {"a": 3}


def test_scripts():
    with tempfile.TemporaryDirectory() as tmpdir:
        params = [{"a": 1}, {"a": 2}, {"a": 3}, {"a": 4}]
        calc_dir = f"{tmpdir}/calc"
        ps.run(func_a, params, calc_dir=calc_dir)

        db = pj(calc_dir, "database.pk")
        print(system(f"psweep-db2json -o columns {db}"))
        print(system(f"psweep-db2table -i -a -f simple {db}"))


def test_backup():
    def func(pset):
        # write stuff to calc_dir
        dr = pj(pset["_calc_dir"], pset["_pset_id"])
        ps.makedirs(dr)
        fn = pj(dr, "foo")
        with open(fn, "w") as fd:
            fd.write(pset["_pset_id"])
        return {"result": pset["a"] * 10}

    with tempfile.TemporaryDirectory() as tmpdir:
        calc_dir = pj(tmpdir, "calc")
        params = [{"a": 1}, {"a": 2}, {"a": 3}, {"a": 4}]

        # First run. backup does nothing yet
        df0 = ps.run(func, params, calc_dir=calc_dir)
        unq = df0._run_id.unique()
        assert len(unq) == 1
        run_id_0 = unq[0]

        # Second run. This time, test backup.
        df1 = ps.run(func, params, calc_dir=calc_dir, backup=True)
        rex = re.compile(r"calc.bak_[0-9-]+T[0-9:\.]+Z_run_id.+")
        found = False
        files = os.listdir(tmpdir)
        for name in os.listdir(tmpdir):
            if rex.search(name) is not None:
                backup_dir = pj(tmpdir, name)
                found = True
                break
        assert found, f"backup dir matching {rex} not found in:\n{files}"
        print(os.listdir(backup_dir))
        msk = df1._run_id == run_id_0
        assert len(df1[msk]) == len(df0)
        assert len(df1) == 2 * len(df0)
        assert len(df1._run_id.unique()) == 2
        for pset_id in df1[msk]._pset_id:
            tgt = pj(backup_dir, pset_id, "foo")
            assert os.path.exists(tgt)
            with open(tgt) as fd:
                fd.read().strip() == pset_id
        assert os.path.exists(pj(backup_dir, "database.pk"))


def test_pass_df_interactive():
    def df_cmp(dfa, dfb):
        assert (dfa.a.values == dfb.a.values).all()
        assert (dfa.result.values == dfb.result.values).all()
        assert (dfa._pset_seq.values == dfb._pset_seq.values).all()
        assert (dfa._run_seq.values == dfb._run_seq.values).all()
        assert (dfa._pset_hash.values == dfb._pset_hash.values).all()

    with tempfile.TemporaryDirectory() as tmpdir:
        calc_dir = pj(tmpdir, "calc")
        params = ps.plist("a", [1, 2, 3, 4])

        # no db disk write for now, test passing in no df, df=None and empty df
        df1_1 = ps.run(func_a, params, calc_dir=calc_dir, save=False)
        df1_2 = ps.run(func_a, params, calc_dir=calc_dir, save=False, df=None)
        df_cmp(df1_1, df1_2)
        df1_3 = ps.run(
            func_a, params, calc_dir=calc_dir, save=False, df=pd.DataFrame()
        )
        df_cmp(df1_1, df1_3)

        # still no disk write, pass in df1 and extend
        df1 = df1_3
        df2 = ps.run(func_a, params, calc_dir=calc_dir, save=False, df=df1)
        assert not os.path.exists(pj(tmpdir, "calc"))
        assert len(df2) == 2 * len(df1)
        assert (df2.a.values == np.tile(df1.a.values, 2)).all()
        assert (
            df2._pset_hash.values == np.tile(df1._pset_hash.values, 2)
        ).all()

        # df2 again, but now write to disk and read (ignore that run()
        # also returns it)
        ps.run(func_a, params, calc_dir=calc_dir, df=df1)
        df2_disk = ps.df_read(pj(calc_dir, "database.pk"))
        df_cmp(df2, df2_disk)


def test_df_filter_conds():
    df = pd.DataFrame({"a": np.arange(10), "b": np.arange(10) + 4})

    # logical and
    conds = [df.a > 3, df.b < 9, df.a % 2 == 0]
    df_ref = df[conds[0] & conds[1] & conds[2]]
    assert df_ref.equals(ps.df_filter_conds(df, conds))
    for filt in [lambda x: x, lambda x: x.to_numpy(), lambda x: x.to_list()]:
        # use iterator map(...)
        assert df_ref.equals(ps.df_filter_conds(df, map(filt, conds)))

    # logical or. For the conds above, OR-ing them happens to result in the
    # original df w/o and reduction.
    assert df.equals(ps.df_filter_conds(df, conds, op="or"))


class _Foo:
    """Dummy custom type used in test_dotdict(). As always, must be defined
    outside b/c Python can't pickle stuff defined in the same scope.
    """

    x = 55

    def __eq__(self, other):
        return self.x == other.x

    def __hash__(self):
        return self.x


# Uncomment to run test_dotdict using one of those implementations.

# https://github.com/cdgriffith/Box
# pip install python-box
##from box import Box
##ps.dotdict = Box

# https://github.com/drgrib/dotmap
# pip install dotmap
##from dotmap import DotMap
##ps.dotdict = DotMap


@pytest.mark.skipif(
    not hasattr(ps, "dotdict"), reason="psweep.dotdict not defined"
)
def test_dotdict():
    """Test a possible implementation for a dict with optional dot attr access.
    See e.g. [1]  for all the stunts people have pulled so far (including
    ourselves [2]!!). The idea is to eventually use that for

    * psets instead of plain dicts
    * general purpose dot access container type (e.g. for param study driver
      scripts)

    None of the present hacks meets all our needs

    * optional dot access
    * update __repr__ after attr setting (e.g. d=dotdict(...); d.x=4)
    * can be pickled

    except for those that live in an external package (box, dotmap), but
    getting some random pypi package as a dep is not worth it. Maybe we'll
    stumble upon a small and neat implementation that we can add to psweep.
    Until then, just stick to plain dicts.


    [1] https://stackoverflow.com/a/23689767
    [2] https://github.com/elcorto/pwtools/blob/master/pwtools/batch.py#L735
    """
    # psweep.py::
    #
    #     class dotdict(dict):
    #         """Not good enough, can't be pickled. So sad!"""
    #         __getattr__ = dict.get
    #         __setattr__ = dict.__setitem__
    #         __delattr__ = dict.__delitem__
    #         __dir__ = dict.keys

    f = _Foo()
    ref = dict(a=1, b=2, c=np.sin, d=f, e=_Foo)

    def dct_cmp(ref, ddict):
        assert ref == ddict
        assert set(ref.keys()) == set(ddict.keys())
        assert set(ref.values()) == set(ddict.values())
        if isinstance(ddict, ps.dotdict):
            ddict.a
            ddict.b
            ddict.c
            ddict.d
            ddict.e

    for ddict in [ps.dotdict(ref), ps.dotdict(**ref)]:
        dct_cmp(ref, ddict)
        # Test whether dotdict implementation can be pickled.
        dct_cmp(ref, pickle.loads(pickle.dumps(ddict)))


def test_pset_hash():
    dref = dict(a=1, b=dict(c=2, d=[1, 2, "a"]))
    d1 = dict(b=dict(c=2, d=[1, 2, "a"]), a=1)
    d2 = dict(a=1, b=dict(d=[1, 2, "a"], c=2))

    assert ps.pset_hash(dref) == "c41c669212f58293fb88aba643b24a995d9aff1a"
    assert ps.pset_hash(d1) == ps.pset_hash(dref)
    assert ps.pset_hash(d2) == ps.pset_hash(dref)

    assert (
        ps.pset_hash(dict(a=np.sin))
        == "32522716b0514eb8b2677a917888c4ca21f792da"
    )

    # These are the correct hashes (tested: ipython or run this function in a
    # script). But pytest's deep introspection "magic" must do something to the
    # _Foo class such that its hash changes, thank you very much. So we just
    # test that hashing them works.
    #
    ##assert (
    ##    ps.pset_hash(dict(a=_Foo))
    ##    == "f7a936832c167d5dc1c9574b937822ad0853c00d"
    ##)
    ##assert (
    ##    ps.pset_hash(dict(a=_Foo()))
    ##    == "7c78a7c995e1cf7a2f1f68f29090b173c9b930fa"
    ##)
    ps.pset_hash(dict(a=_Foo))
    ps.pset_hash(dict(a=_Foo()))

    d_no_us = dict(a=1, b=2)
    d_us = dict(a=1, b=2, _c=3)
    assert ps.pset_hash(d_no_us) == ps.pset_hash(d_us)
    assert ps.pset_hash(d_no_us) == ps.pset_hash(d_us, skip_special_cols=True)
    assert not ps.pset_hash(d_no_us) == ps.pset_hash(
        d_us, skip_special_cols=False
    )


def test_param_build():
    a = ps.plist("a", [1, 2])
    b = ps.plist("b", [77, 88, 99])

    assert a == ps.pgrid([a])
    with pytest.raises(AssertionError):
        ps.pgrid(a)

    abref = [
        {"a": 1, "b": 77},
        {"a": 1, "b": 88},
        {"a": 1, "b": 99},
        {"a": 2, "b": 77},
        {"a": 2, "b": 88},
        {"a": 2, "b": 99},
    ]

    assert abref == ps.pgrid(a, b)
    assert abref == ps.pgrid([a, b])
    assert abref == ps.itr2params(product(a, b))

    # a and c together, no product() here so it is an error to use pgrid(a,c)
    # which calls product(): pgrid(a,c) = itr2params(product(a,c)). We have
    # only
    #
    # >>> list(zip(a,c))
    # [({'a': 1}, {'c': 11}),
    #  ({'a': 2}, {'c': 22})]
    # >>> ps.itr2params(zip(a,c))
    # [{'a': 1, 'c': 11},
    #  {'a': 2, 'c': 22}]
    c = ps.plist("c", [11, 22])
    acref = [{"a": 1, "c": 11}, {"a": 2, "c": 22}]
    assert acref == ps.itr2params(zip(a, c))
    assert acref == ps.pgrid([zip(a, c)])
    with pytest.raises(AssertionError):
        ps.pgrid(zip(a, c))

    d = ps.plist("d", [66, 77, 88, 99])
    acdref = ps.itr2params(product(zip(a, c), d))
    assert acdref == ps.pgrid([zip(a, c), d])
    assert acdref == ps.pgrid(zip(a, c), d)


def test_itr():
    @ps.itr
    def func(args):
        return [a for a in args]

    def gen(inp):
        for x in inp:
            yield x

    assert [1] == func(1)
    assert [1] == func([1])
    assert [[1]] == func([[1]])
    assert [1, 2] == func(1, 2)
    assert [1, 2] == func([1, 2])
    assert [[1, 2]] == func([[1, 2]])
    assert [1, 2] == func(gen([1, 2]))


def test_filter_params_unique():
    params = [
        dict(a=1, b=2, c=3),  # same!
        dict(a=1, b=2, c=3),  # same!
        dict(a=1, b=2, c=444),
    ]

    params_filt = ps.filter_params_unique(params)
    assert len(params_filt) == 2
    for idx in [0, 2]:
        assert params[idx] in params_filt


def test_filter_params_dup_hash():
    params = [
        dict(a=1, b=2, c=3),  # same!
        dict(a=1, b=2, c=3),  # same!
        dict(a=1, b=2, c=333),
        dict(a=1, b=2, c=444),
    ]

    hashes = [ps.pset_hash(pset) for pset in params]
    params_filt = ps.filter_params_dup_hash(params, hashes)
    assert len(params_filt) == 0

    hashes = [ps.pset_hash(pset) for pset in params[:3]]
    params_filt = ps.filter_params_dup_hash(params, hashes)
    assert len(params_filt) == 1
    assert params_filt[0] == params[-1]


def test_stargrid():
    a = ps.plist("a", [1, 2, 3])
    b = ps.plist("b", [77, 88, 99])
    c = ps.plist("c", [np.sin, np.cos, _Foo, _Foo()])
    const = dict(a=1, b=77)

    # API
    ps.stargrid(const, [a, b])
    ps.stargrid(const=const, vary=[a, b])

    params = ps.stargrid(const, [a, b], skip_dups=False)
    assert len(params) == 6
    params = ps.stargrid(const, [a, b])
    assert len(params) == 5
    params = ps.stargrid(const, [c])
    assert len(params) == len(c)
    params = ps.stargrid(dict(), [a])
    assert params == a

    vary_labels = ["aa", "bb"]
    params = ps.stargrid(const, [a, b], vary_labels=vary_labels)
    for pset in params:
        assert "_vary" in pset.keys()
        assert pset["_vary"] in vary_labels

    params = ps.stargrid(
        const, [a, b], vary_labels=vary_labels, vary_label_col="study"
    )
    for pset in params:
        assert "study" in pset.keys()
        assert "_vary" not in pset.keys()
        assert pset["study"] in vary_labels


def test_intspace():
    assert (ps.intspace(0, 4, 5) == np.array(range(5))).all()
    # array([0.  , 1.25, 2.5 , 3.75, 5.  ]) -> round to int
    assert (ps.intspace(0, 5, 5) == np.array([0, 1, 2, 4, 5])).all()
    # array([0, 1, 2, 3, 4, 5])
    assert (ps.intspace(0, 5, 20) == np.array(range(6))).all()


def test_logspace_defaults():
    start = np.random.rand()
    stop = 5 * start
    log_func = np.log10
    ref = np.logspace(log_func(start), log_func(stop))
    np.testing.assert_allclose(ref, ps.logspace(start, stop))


@pytest.mark.parametrize("log_func, base", [(np.log10, 10), (np.log, np.e)])
def test_logspace_base_log_func(log_func, base):
    start = np.random.rand()
    stop = 5 * start
    ref = np.logspace(log_func(start), log_func(stop), base=base, num=20)
    # offset=0
    np.testing.assert_allclose(
        ref, ps.logspace(start, stop, num=20, base=base, log_func=log_func)
    )


@pytest.mark.parametrize("offset", [1, 2.345])
def test_logspace_offset(offset):
    start = np.random.rand()
    stop = 5 * start
    log_func = np.log10
    ref = np.logspace(log_func(start), log_func(stop))
    val = ps.logspace(start, stop, offset=offset)
    np.testing.assert_allclose(val[0], start)
    np.testing.assert_allclose(val[-1], stop)
    assert not np.allclose(ref, val)


def test_prep_batch():
    a = ps.plist("param_a", [1, 2, 3])
    b = ps.plist("param_b", ["xx", "yy"])
    params = ps.pgrid(a, b)
    template_dir = f"{here}/../examples/batch_templates_git/templates"
    with tempfile.TemporaryDirectory() as tmpdir:
        # run 0
        df = ps.prep_batch(
            params,
            calc_dir=f"{tmpdir}/calc",
            calc_templ_dir=f"{template_dir}/calc",
            machine_templ_dir=f"{template_dir}/machines",
            database_dir=f"{tmpdir}/db",
            write_pset=True,
        )
        print(system(f"tree {tmpdir}"))

        assert os.path.exists(f"{tmpdir}/db/database.pk")
        assert os.path.exists(f"{tmpdir}/calc/run_local.sh")
        assert os.path.exists(f"{tmpdir}/calc/run_cluster.sh")

        n_param = len(params)
        for txt in [
            ps.file_read(f"{tmpdir}/calc/run_cluster.sh"),
            ps.file_read(f"{tmpdir}/calc/run_local.sh"),
        ]:
            assert "# run_seq=0 pset_seq=0" in txt
            assert f"# run_seq=0 pset_seq={n_param - 1}" in txt

        for pset_id in df._pset_id.values:
            for name in [
                "run.py",
                "pset.pk",
                "jobscript_local",
                "jobscript_cluster",
            ]:
                assert os.path.exists(f"{tmpdir}/calc/{pset_id}/{name}")

        # run 1
        df = ps.prep_batch(
            params,
            calc_dir=f"{tmpdir}/calc",
            calc_templ_dir=f"{template_dir}/calc",
            machine_templ_dir=f"{template_dir}/machines",
            database_dir=f"{tmpdir}/db",
            write_pset=True,
        )

        for txt in [
            ps.file_read(f"{tmpdir}/calc/run_cluster.sh"),
            ps.file_read(f"{tmpdir}/calc/run_local.sh"),
        ]:
            assert "# run_seq=0 pset_seq=0" in txt
            assert f"# run_seq=0 pset_seq={n_param - 1}" in txt
            assert f"# run_seq=1 pset_seq={n_param}" in txt
            assert f"# run_seq=1 pset_seq={2*n_param - 1}" in txt


# We can't use
#
#   @pytest.mark.skipif(
#       importlib.util.find_spec("dask.distributed") is None,
#       reason="dask.distributed not found",
#   )
#
# If both dask and distributed (import as dask.distributed) are *not*
# installed, find_spec("dask.distributed") errors out saying that dask is not
# installed. This is because (from the docs):
#
#   importlib.util.find_spec(name, package=None)
#
#   If name is for a submodule (contains a dot), the parent module is
#   automatically imported.
#
# OK, so instead of testing if dask is installed, it just tries to import it.
#
# We can't solve this by using a second skipif to test for dask first.
#
#   @pytest.mark.skipif(
#       importlib.util.find_spec("dask") is None,
#       reason="dask not found",
#   )
#   @pytest.mark.skipif(
#       importlib.util.find_spec("dask.distributed") is None,
#       reason="dask.distributed not found",
#   )
#
# This doesn't work because pytest seems to always test both skipifs, no matter
# which wrapper comes first, so the order is irrelevant.
#
# The only solution we have so far is the one below. Since "foo or bar" is
# evaluated left-to-right, we first check if dask is installed and stop if not.
#
@pytest.mark.skipif(
    (importlib.util.find_spec("dask") is None)
    or (importlib.util.find_spec("dask.distributed") is None),
    reason="dask or dask.distributed not found",
)
def test_dask_local_cluster():
    from dask.distributed import Client, LocalCluster

    # LocalCluster is default if no cluster is provided
    client = Client(dashboard_address=None)

    params = ps.plist("a", [1, 2, 3])
    df = ps.run(func_a, params, dask_client=client, save=False)
    assert len(df) == 3
    client.close()

    # We don't need the dask dashboard in tests. Setting dashboard_address=None
    # should turn the dashboard off. But that doesn't work
    # (https://github.com/dask/distributed/issues/8136). We see warnings like
    #   UserWarning: Port 8787 is already in use. Perhaps you already have a
    #   cluster running? ...
    # as a result, if we run tests in parallel.
    cluster = LocalCluster(dashboard_address=None)
    client = Client(cluster)

    df = ps.run(func_a, params, dask_client=client, save=False, df=df)
    assert len(df) == 6


def test_pickle_io():
    obj = dict(a=1, b=_Foo(), c=np.sin)
    hsh = lambda obj: joblib.hash(obj, hash_name="sha1")
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = f"{tmpdir}/path/that/has/to/be/created/file.pk"
        ps.pickle_write(fn, obj)
        assert hsh(obj) == hsh(ps.pickle_read(fn))


def test_file_io():
    txt = "some random text"
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = f"{tmpdir}/path/that/has/to/be/created/file.txt"
        ps.file_write(fn, txt)
        assert txt == ps.file_read(fn)


def test_single_uuid():
    existing = [str(uuid.uuid4()) for _ in range(100)]
    ret = ps.get_uuid()
    # 'b5233ba3-25eb-40ac-b2b1-d1babbacd904'
    assert isinstance(ret, str)
    assert len(ret) == 36
    assert re.match("([a-z0-9]+-){4}[a-z0-9]", ret) is not None
    ret = ps.get_uuid(existing=existing)
    assert ret not in existing


def test_many_uuids():
    existing = [str(uuid.uuid4()) for _ in range(100)]
    ret = set(ps.get_many_uuids(100))
    assert len(ret) == 100

    ret = set(ps.get_many_uuids(100, existing=existing))
    assert len(ret) == 100
    assert len(ret & set(existing)) == 0


def test_run_not_alter_params():
    params = ps.plist("a", [1, 2, 3])
    params_copy = copy.deepcopy(params)
    hsh = joblib.hash(params, hash_name="sha1")
    df = ps.run(func_a, params, save=False)
    df = ps.run(func_a, params, save=False, df=df)
    assert params == params_copy
    assert hsh == joblib.hash(params, hash_name="sha1")
    for pset in params:
        for key in pset.keys():
            assert not key.startswith("_")


class TestCaptureLogs:
    def get_client(self, use_dask):
        if use_dask:
            from dask.distributed import Client, LocalCluster

            cluster = LocalCluster(dashboard_address=None)
            cluster.scale(n=3)
            client = Client(cluster)
            return client
        else:
            return None

    @staticmethod
    def get_print_funcs():
        fout = lambda msg: sys.stdout.write(msg)
        ferr = lambda msg: sys.stderr.write(msg)
        return fout, ferr

    @staticmethod
    def get_print_funcs_fail(sout=sys.stdout, serr=sys.stderr):
        """This case will fail, i.e. printed text will disappear into the
        nirvana. The reason is that the redirect machinery in run() will
        dynamically redirect sys.sydout and sys.stderr at runtime. However
        here, we have bound those streams to the sout and serr variables at
        parse time before run() can touch them, so they won't be affected.

        This is an example of user code where the redirect feature cannot be
        used.
        """
        fout = lambda msg: sout.write(msg)
        ferr = lambda msg: serr.write(msg)
        return fout, ferr

    def func_with_logs(self, pset, f_get_print_funcs=None):
        print(f"{pset=}")
        print("txt on stdout")
        print("txt on stderr", file=sys.stderr)
        sys.stdout.write("stdout direct")
        sys.stderr.write("stderr direct")
        fout, ferr = f_get_print_funcs()
        fout("txt from fout")
        ferr("txt from ferr")
        return dict()

    @staticmethod
    def assert_txt_content(txt):
        assert "pset=" in txt
        assert "txt on stdout" in txt
        assert "txt on stderr" in txt
        assert "stderr direct" in txt
        assert "stdout direct" in txt
        assert "txt from fout" in txt
        assert "txt from ferr" in txt

    def assert_log_content(self, df, run_id, in_file=False, in_db=False):
        dfs = df[df._run_id == run_id]
        assert len(dfs) == 3
        for calc_dir, pset_id in zip(
            dfs._calc_dir.values, dfs._pset_id.values
        ):
            if in_db:
                db_txt = dfs.loc[dfs._pset_id == pset_id, "_logs"].iloc[0]
                assert len(db_txt) > 0
                self.assert_txt_content(db_txt)
            fn = pj(calc_dir, pset_id, "logs.txt")
            if in_file:
                assert os.path.exists(fn)
                txt = ps.file_read(fn)
                assert len(txt) > 0
                self.assert_txt_content(txt)
                if in_db:
                    assert db_txt == txt
            else:
                assert not os.path.exists(fn)

    @staticmethod
    def get_parametrize_mark(test_func):
        xfail = pytest.mark.xfail(raises=AssertionError, strict=True)
        wrapper = pytest.mark.parametrize(
            "should_pass", [True, pytest.param(False, marks=xfail)]
        )
        return wrapper(test_func)

    @get_parametrize_mark
    def test(self, should_pass):
        self.impl(use_dask=False, should_pass=should_pass)

    @pytest.mark.skipif(
        (importlib.util.find_spec("dask") is None)
        or (importlib.util.find_spec("dask.distributed") is None),
        reason="dask or dask.distributed not found",
    )
    @get_parametrize_mark
    def test_with_dask(self, should_pass):
        self.impl(use_dask=True, should_pass=should_pass)

    def impl(self, use_dask, should_pass):
        client = self.get_client(use_dask)
        func = partial(
            self.func_with_logs,
            f_get_print_funcs=self.get_print_funcs
            if should_pass
            else self.get_print_funcs_fail,
        )
        with tempfile.TemporaryDirectory() as tmpdir:
            calc_dir = f"{tmpdir}/calc"
            params = ps.plist("a", [1, 2, 3])
            df = ps.run(
                func,
                params,
                calc_dir=calc_dir,
                capture_logs="file",
                dask_client=client,
            )
            assert not hasattr(df, "_logs")
            self.assert_log_content(
                df, run_id=df._run_id.values[-1], in_file=True
            )

            df = ps.run(
                func,
                params,
                calc_dir=calc_dir,
                capture_logs="db",
                dask_client=client,
            )
            assert hasattr(df, "_logs")
            self.assert_log_content(
                df, run_id=df._run_id.values[-1], in_db=True
            )

            df = ps.run(
                func,
                params,
                calc_dir=calc_dir,
                capture_logs="db+file",
                dask_client=client,
            )
            assert hasattr(df, "_logs")
            self.assert_log_content(
                df, run_id=df._run_id.values[-1], in_file=True, in_db=True
            )


# We don't go the extra mile and verify the text output, but you may check like
# so:
#
# $ pytest -k verbose -vs
# tests/test_all.py::test_verbose[True]
#                                a
# 2023-11-22 19:41:49.736106634  1
#                                a
# 2023-11-22 19:41:49.739609957  2
#                                a
# 2023-11-22 19:41:49.742296457  3
# PASSED
# tests/test_all.py::test_verbose[False] PASSED
# tests/test_all.py::test_verbose[verbose2]
#                                                            _pset_id  a
# 2023-11-22 19:41:49.752099991  38a6820f-7e41-40af-b403-3647f44e8b9d  1
#                                                            _pset_id  a
# 2023-11-22 19:41:49.755405426  9ade6afb-1fd3-4d29-9e7a-0bacdb950994  2
#                                                            _pset_id  a
# 2023-11-22 19:41:49.757858753  80193006-56d4-47b1-be95-fe06cee2eec1  3
# PASSED
@pytest.mark.parametrize("verbose", [True, False, ["a", "_pset_id"]])
def test_verbose(verbose):
    params = ps.plist("a", [1, 2, 3])
    ps.run(func_a, params, save=False, verbose=verbose)



from functools import partial, wraps
from io import IOBase, StringIO
from typing import Any, Sequence, Callable, Iterator
import copy
from contextlib import redirect_stdout, redirect_stderr
import itertools
import multiprocessing as mp
import os
import pickle
import platform
import re
import shutil
import string
import subprocess
import sys
import time
import uuid
import warnings

import joblib
import numpy as np
import pandas as pd
import yaml

pj = os.path.join

# defaults, globals
PANDAS_DEFAULT_ORIENT = "records"
PANDAS_TIME_UNIT = "s"
PSET_HASH_ALG = "sha1"
GIT_ADD_ALL = "git add -A -v"

# Make DeprecationWarning visible to users by default.
warnings.simplefilter("default")


# -----------------------------------------------------------------------------
# helpers
# -----------------------------------------------------------------------------


def system(cmd: str, **kwds) -> subprocess.CompletedProcess:
    """
    Call shell command.

    Parameters
    ----------
    cmd
        shell command
    kwds
        keywords passed to `subprocess.run`
    """
    try:
        return subprocess.run(
            cmd,
            shell=True,
            check=True,
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
            **kwds,
        )
    except subprocess.CalledProcessError as ex:
        print(ex.stdout.decode())
        raise ex


# https://github.com/elcorto/pwtools
def makedirs(path: str) -> None:
    """Create `path` recursively, no questions asked."""
    if not path.strip() == "":
        os.makedirs(path, exist_ok=True)


# https://github.com/elcorto/pwtools
def fullpath(path: str) -> str:
    return os.path.abspath(os.path.expanduser(path))


def itr(func: Callable) -> Callable:
    """Wrap `func` to allow pasing args not as sequence.

    Assuming ``func()`` requires a sequence as input: ``func([a,b,c])``, allow
    passing ``func(a,b,c)``.
    """

    @wraps(func)
    def wrapper(*args):
        # (arg1,)
        if len(args) == 1:
            arg = args[0]
            return func(arg if is_seq(arg) else [arg])
        # (arg1,...,argN)
        else:
            return func(args)

    return wrapper


# https://github.com/elcorto/pwtools
def is_seq(seq) -> bool:
    if (
        isinstance(seq, str)
        or isinstance(seq, IOBase)
        or isinstance(seq, dict)
    ):
        return False
    else:
        try:
            iter(seq)
            return True
        except TypeError:
            return False


def flatten(seq):
    for item in seq:
        if not is_seq(item):
            yield item
        else:
            for subitem in flatten(item):
                yield subitem


def file_write(fn: str, txt: str, mode="w"):
    makedirs(os.path.dirname(fn))
    with open(fn, mode=mode) as fd:
        fd.write(txt)


def file_read(fn: str):
    with open(fn, "r") as fd:
        return fd.read()


def pickle_write(fn: str, obj):
    makedirs(os.path.dirname(fn))
    with open(fn, "wb") as fd:
        pickle.dump(obj, fd)


def pickle_read(fn: str):
    with open(fn, "rb") as fd:
        return pickle.load(fd)


class PsweepHashError(TypeError):
    pass


def pset_hash(
    dct: dict,
    method=PSET_HASH_ALG,
    raise_error=True,
    skip_special_cols=True,
):
    """Reproducible hash of a dict for usage in database (hash of a `pset`)."""

    # We target "reproducible" hashes, i.e. not what Python's ``hash`` function
    # does, for instance for two interpreter sessions::
    #
    #     $ python
    #     >>> hash("12")
    #     8013944793133897043
    #
    #     $ python
    #     >>> hash("12")
    #     4021864388667373027
    #
    # We can't hash, say, the pickled byte string of some object (e.g.
    # ``hash(pickle.dumps(obj))``), b/c that may contain a ref to its memory
    # location which is not what we're interested in. Similarly, also using
    # ``repr`` is not reproducible::
    #
    #     >>> class Foo:
    #     ...     pass
    #
    #     >>> repr(Foo())
    #     '<__main__.Foo object at 0x7fcc68aa9d60>'
    #     >>> repr(Foo())
    #     '<__main__.Foo object at 0x7fcc732034c0>'
    #
    # even though for our purpose, we'd consider the two instances of ``Foo``
    # to be the same.
    #
    # The same observations have been also made elsewhere [1,2]. Esp. [2]
    # points to [3] which in turn mentions joblib.hashing.hash(). It's code
    # shows how complex the problem is, but so far this is our best bet.
    #
    # [1] https://death.andgravity.com/stable-hashing
    # [2] https://ourpython.com/python/deterministic-recursive-hashing-in-python
    # [3] https://stackoverflow.com/a/52175075
    if skip_special_cols:
        _dct = {
            key: val for key, val in dct.items() if not key.startswith("_")
        }
    else:
        _dct = dct
    # joblib can hash "anything" so we didn't come up with an input that
    # actually fails to hash. As such, TypeError is just a guess here. But
    # still we don't catch ValueError raised when an invalid hash_name is
    # passed (anything other than md5 or sha1).
    try:
        return joblib.hash(_dct, hash_name=method)
    except TypeError as ex:
        if raise_error:
            raise PsweepHashError(
                f"Error in hash calculation of: {dct}"
            ) from ex
        else:
            return np.nan


def check_calc_dir(calc_dir: str, df: pd.DataFrame):
    """Check calc dir for consistency with database.

    Assuming dirs are named::

        <calc_dir>/<pset_id1>
        <calc_dir>/<pset_id2>
        ...

    check if we have matching dirs to ``_pset_id`` values in the
    database.
    """
    # fmt: off
    pset_ids_disk = set([
        m.group()
        for m in [
            re.match(r"(([0-9a-z]+)-){4}([0-9a-z]+)", x)
            for x in os.listdir(calc_dir)
        ] if m is not None])
    # fmt: on
    pset_ids_db = set(df._pset_id.values)
    return dict(
        db_not_disk=pset_ids_db - pset_ids_disk,
        disk_not_db=pset_ids_disk - pset_ids_db,
    )


def logspace(
    start, stop, num=50, offset=0, log_func: Callable = np.log10, **kwds
):
    """
    Like ``numpy.logspace`` but

    * `start` and `stop` are not exponents but the actual bounds
    * tuneable log scale strength

    Control the strength of the log scale by `offset`, where we use by default
    ``log_func=np.log10`` and ``base=10`` and return
    ``np.logspace(np.log10(start + offset), np.log10(stop + offset)) -
    offset``. `offset=0` is equal to ``np.logspace(np.log10(start),
    np.log10(stop))``. Higher `offset` values result in more evenly spaced
    points.

    Parameters
    ----------
    start, stop, num, **kwds :
        same as in ``np.logspace``
    offset :
        Control strength of log scale.
    log_func :
        Must match `base` (pass that as part of `**kwds`). Default is
        ``base=10`` as in ``np.logspace`` and so ``log_func=np.log10``. If you
        want a different `base`, also provide a matching `log_func`, e.g.
        ``base=e, log_func=np.log``.

    Examples
    --------
    Effect of different `offset` values:

    >>> from matplotlib import pyplot as plt
    >>> from psweep import logspace
    >>> import numpy as np
    >>> for ii, offset in enumerate([1e-16,1e-3, 1,2,3]):
    ...     x=logspace(0, 2, 20, offset=offset)
    ...     plt.plot(x, np.ones_like(x)*ii, "o-", label=f"{offset=}")
    >>> plt.legend()
    """
    base = kwds.pop("base", 10.0)
    # fmt: off
    assert np.allclose(log_func(base), 1.0), f"log_func and {base=} don't match"
    # fmt: on
    return (
        np.logspace(
            log_func(start + offset),
            log_func(stop + offset),
            num=num,
            base=base,
            **kwds,
        )
        - offset
    )


def intspace(*args, dtype=np.int64, **kwds):
    """Like ``np.linspace`` but round to integers.

    The length of the returned array may be lower than specified by `num` if
    rounding to ints results in duplicates.

    Parameters
    ----------
    *args, **kwds
        Same as ``np.linspace``
    """
    assert "dtype" not in kwds, "Got 'dtype' multiple times."
    return np.unique(np.round(np.linspace(*args, **kwds)).astype(dtype))


def get_uuid(retry=10, existing: Sequence = []) -> str:
    ret = str(uuid.uuid4())
    while ret in existing:
        ret = str(uuid.uuid4())
        retry -= 1
        if retry == 0:
            raise Exception(
                f"Failed to generate UUID after {retry} attempts. "
                f"Existing UUIDs: {existing}"
            )
    return ret


def get_many_uuids(
    num: int, retry=10, existing: Sequence = []
) -> Sequence[str]:
    generate = lambda: set([str(uuid.uuid4()) for _ in range(num)])
    ret = generate()
    while (len(ret) < num) or (len(ret & set(existing)) > 0):
        ret = generate()
        retry -= 1
        if retry == 0:
            raise Exception(
                f"Failed to generate {num} UUIDs after {retry} attempts. "
                f"Existing UUIDs: {existing}"
            )
    return list(ret)


# -----------------------------------------------------------------------------
# git
# -----------------------------------------------------------------------------


def git_clean():
    return system("git status --porcelain").stdout.decode() == ""


def in_git_repo():
    # fmt: off
    return subprocess.run(
        "git status",
        check=False,
        shell=True,
        capture_output=True,
        ).returncode == 0
    # fmt: on


def git_enter(use_git: bool, always_commit=False):
    if use_git:
        path = os.path.basename(fullpath(os.curdir))
        if not in_git_repo():
            if always_commit:
                system(
                    f"git init; {GIT_ADD_ALL}; git commit -m 'psweep: {path}: init'"
                )
            else:
                raise Exception("no git repo here, create one first")
        if not git_clean():
            if always_commit:
                print("dirty repo, adding all changes")
                system(
                    f"{GIT_ADD_ALL}; git commit -m 'psweep: {path}: local changes'"
                )
            else:
                raise Exception("dirty repo, commit first")


def git_exit(use_git: bool, df: pd.DataFrame):
    if use_git and (not git_clean()):
        path = os.path.basename(fullpath(os.curdir))
        system(
            f"{GIT_ADD_ALL}; git commit -m 'psweep: {path}: run_id={df._run_id.values[-1]}'"
        )


# -----------------------------------------------------------------------------
# pandas
# -----------------------------------------------------------------------------


def df_to_json(df: pd.DataFrame, **kwds) -> str:
    """Like `df.to_json` but with defaults for orient, date_unit, date_format,
    double_precision.

    Parameters
    ----------
    df
        DataFrame to convert
    kwds
        passed to :meth:`df.to_json`
    """
    defaults = dict(
        orient=PANDAS_DEFAULT_ORIENT,
        date_unit=PANDAS_TIME_UNIT,
        date_format="iso",
        double_precision=15,
    )
    for key, val in defaults.items():
        if key not in kwds.keys():
            kwds[key] = val
    return df.to_json(**kwds)


def df_write(fn: str, df: pd.DataFrame, fmt="pickle", **kwds) -> None:
    """Write DataFrame to disk.

    Parameters
    ----------
    fn
        filename
    df
        DataFrame to write
    fmt
        ``{'pickle', 'json'}``
    kwds
        passed to ``pickle.dump()`` or :func:`df_to_json`
    """
    makedirs(os.path.dirname(fn))
    if fmt == "pickle":
        with open(fn, "wb") as fd:
            pickle.dump(df, fd, **kwds)
    elif fmt == "json":
        df_to_json(df, path_or_buf=fn, **kwds)
    else:
        raise ValueError("unknown fmt: {}".format(fmt))


def df_read(fn: str, fmt="pickle", **kwds):
    """Read DataFrame from file `fn`. See :func:`df_write`."""
    if fmt == "pickle":
        with open(fn, "rb") as fd:
            return pickle.load(fd, **kwds)
    elif fmt == "json":
        orient = kwds.pop("orient", PANDAS_DEFAULT_ORIENT)
        return pd.io.json.read_json(
            fn, precise_float=True, orient=orient, **kwds
        )
    else:
        raise ValueError("unknown fmt: {}".format(fmt))


def df_print(
    df: pd.DataFrame,
    index: bool = False,
    special_cols: bool = False,
    cols: Sequence[str] = [],
    skip_cols: Sequence[str] = [],
):
    """Print DataFrame, by default without the index and special columns such
    as `_pset_id`.

    Similar logic as in `bin/psweep-db2table`, w/o tabulate support but more
    features (`skip_cols` for instance).

    Column names are always sorted, so the order of names in e.g. `cols`
    doesn't matter.

    Parameters
    ----------
    df
    index
        include DataFrame index
    special_cols
        include all special columns (`_pset_id` etc.)
    cols
        explicit sequence of columns, overrides `special_cols` when special columns
        are specified
    skip_cols
        skip those columns instead of selecting them (like `cols` would), use
        either this or `cols`; overrides `special_cols` when special columns
        are specified

    Examples
    --------
    >>> import pandas as pd
    >>> df=pd.DataFrame(dict(a=rand(3), b=rand(3), _c=rand(3)))

    >>> df
              a         b        _c
    0  0.373534  0.304302  0.161799
    1  0.698738  0.589642  0.557172
    2  0.343316  0.186595  0.822023

    >>> ps.df_print(df)
           a        b
    0.373534 0.304302
    0.698738 0.589642
    0.343316 0.186595

    >>> ps.df_print(df, special_cols=True)
           a        b       _c
    0.373534 0.304302 0.161799
    0.698738 0.589642 0.557172
    0.343316 0.186595 0.822023

    >>> ps.df_print(df, index=True)
              a        b
    0  0.373534 0.304302
    1  0.698738 0.589642
    2  0.343316 0.186595

    >>> ps.df_print(df, cols=["a"])
           a
    0.373534
    0.698738
    0.343316

    >>> ps.df_print(df, cols=["a"], special_cols=True)
           a       _c
    0.373534 0.161799
    0.698738 0.557172
    0.343316 0.822023

    >>> ps.df_print(df, cols=["a", "_c"])
           a       _c
    0.373534 0.161799
    0.698738 0.557172
    0.343316 0.822023

    >>> ps.df_print(df, skip_cols=["a"])
           b
    0.304302
    0.589642
    0.186595
    """
    _special_cols = set(x for x in df.columns if x.startswith("_"))
    if len(cols) > 0:
        if len(skip_cols) > 0:
            raise ValueError("Use either skip_cols or cols")
        disp_cols = set(cols) | (_special_cols if special_cols else set())
    else:
        disp_cols = set(df.columns) - (
            set() if special_cols else _special_cols
        )
        if len(skip_cols) > 0:
            disp_cols = disp_cols - set(skip_cols)
    disp_cols = list(disp_cols)
    disp_cols.sort()
    print(df[disp_cols].to_string(index=index))


def df_filter_conds(
    df: pd.DataFrame,
    conds: Sequence[Sequence[bool]],
    op: str = "and",
) -> pd.DataFrame:
    """Filter DataFrame using bool arrays/Series/DataFrames in `conds`.

    Fuse all bool sequences in `conds` using `op`. For instance, if
    ``op="and"``, then we logical-and them, which is equal to

    >>> df[conds[0] & conds[1] & conds[2] & ...]

    but `conds` can be programmatically generated while the expression above
    would need to be changed by hand if `conds` changes.

    Parameters
    ----------
    df
        DataFrame
    conds
        Sequence of bool masks, each of length `len(df)`.
    op
        Bool operator, used as ``numpy.logical_{op}``, e.g. "and", "or",
        "xor".

    Returns
    -------
    DataFrame

    Examples
    --------
    >>> df=pd.DataFrame({'a': arange(10), 'b': arange(10)+4})
    >>> c1=df.a > 3
    >>> c2=df.b < 9
    >>> c3=df.a % 2 == 0
    >>> df[c1 & c2 & c3]
       a  b
    4  4  8
    >>> ps.df_filter_conds(df, [c1, c2, c3])
       a  b
    4  4  8
    """
    cc = conds if hasattr(conds, "__len__") else list(conds)
    if len(cc) == 0:
        return df
    for ic, c in enumerate(cc):
        # fmt: off
        assert len(c) == len(df), \
            f"Condition at index {ic} has {len(c)=}, expect {len(df)=}"
        # fmt: on
    if len(cc) == 1:
        msk = cc[0]
    else:
        assert op in (
            op_allowed := ["and", "or", "xor"]
        ), f"{op=} not one of {op_allowed}"
        msk = getattr(np, f"logical_{op}").reduce(cc)
    return df[msk]


# -----------------------------------------------------------------------------
# building params
# -----------------------------------------------------------------------------


def plist(name: str, seq: Sequence[Any]):
    """Create a list of single-item dicts holding the parameter name and a
    value.

    >>> plist('a', [1,2,3])
    [{'a': 1}, {'a': 2}, {'a': 3}]
    """
    return [{name: entry} for entry in seq]


@itr
def merge_dicts(args: Sequence[dict]):
    """Start with an empty dict and update with each arg dict
    left-to-right."""
    dct = {}
    assert is_seq(args), f"input {args=} is no sequence"
    for entry in args:
        assert isinstance(entry, dict), f"{entry=} is no dict"
        dct.update(entry)
    return dct


def itr2params(loops: Iterator[Any]):
    """Transform the (possibly nested) result of a loop over plists (or
    whatever has been used to create psets) to a proper list of psets
    by flattening and merging dicts.

    Examples
    --------
    >>> a = ps.plist('a', [1,2])
    >>> b = ps.plist('b', [77,88])
    >>> c = ps.plist('c', ['const'])

    >>> # result of loops
    >>> list(itertools.product(a,b,c))
    [({'a': 1}, {'b': 77}, {'c': 'const'}),
     ({'a': 1}, {'b': 88}, {'c': 'const'}),
     ({'a': 2}, {'b': 77}, {'c': 'const'}),
     ({'a': 2}, {'b': 88}, {'c': 'const'})]

    >>> # flatten into list of psets
    >>> ps.itr2params(itertools.product(a,b,c))
    [{'a': 1, 'b': 77, 'c': 'const'},
     {'a': 1, 'b': 88, 'c': 'const'},
     {'a': 2, 'b': 77, 'c': 'const'},
     {'a': 2, 'b': 88, 'c': 'const'}]

    >>> # also more nested stuff is no problem
    >>> list(itertools.product(zip(a,b),c))
    [(({'a': 1}, {'b': 77}), {'c': 'const'}),
     (({'a': 2}, {'b': 88}), {'c': 'const'})]

    >>> ps.itr2params(itertools.product(zip(a,b),c))
    [{'a': 1, 'b': 77, 'c': 'const'},
     {'a': 2, 'b': 88, 'c': 'const'}]
    """
    ret = [merge_dicts(flatten(entry)) for entry in loops]
    lens = list(map(len, ret))
    assert (
        len(np.unique(lens)) == 1
    ), f"not all psets have same length {lens=}\n  {ret=}"
    return ret


@itr
def pgrid(plists: Sequence[Sequence[dict]]) -> Sequence[dict]:
    """Convenience function for the most common loop: nested loops with
    ``itertools.product``: ``ps.itr2params(itertools.product(a,b,c,...))``.

    Parameters
    ----------
    plists
        List of :func:`plist()` results. If more than one, you can also provide
        plists as args, so ``pgrid(a,b,c)`` instead of ``pgrid([a,b,c])``.

    Notes
    -----
    For a single plist arg, you have to use ``pgrid([a])``. ``pgrid(a)`` won't
    work. However, this edge case (passing one plist to pgrid) is not super
    useful, since

    >>> a=ps.plist("a", [1,2,3])
    >>> a
    [{'a': 1}, {'a': 2}, {'a': 3}]
    >>> ps.pgrid([a])
    [{'a': 1}, {'a': 2}, {'a': 3}]

    Examples
    --------
    >>> a = ps.plist('a', [1,2])
    >>> b = ps.plist('b', [77,88])
    >>> c = ps.plist('c', ['const'])
    >>> # same as pgrid([a,b,c])
    >>> ps.pgrid(a,b,c)
    [{'a': 1, 'b': 77, 'c': 'const'},
     {'a': 1, 'b': 88, 'c': 'const'},
     {'a': 2, 'b': 77, 'c': 'const'},
     {'a': 2, 'b': 88, 'c': 'const'}]

    >>> ps.pgrid(zip(a,b),c)
    [{'a': 1, 'b': 77, 'c': 'const'},
     {'a': 2, 'b': 88, 'c': 'const'}]
    """
    assert is_seq(plists), f"input {plists=} is no sequence"
    return itr2params(itertools.product(*plists))


def filter_params_unique(params: Sequence[dict], **kwds) -> Sequence[dict]:
    """Reduce params to unique psets.

    Use pset["_pset_hash"]  if present, else calculate hash on the fly.

    Parameters
    ----------
    params
    kwds
        passed to :func:`pset_hash`
    """
    get_hash = lambda pset: pset.get("_pset_hash", pset_hash(pset, **kwds))
    msk = np.unique([get_hash(pset) for pset in params], return_index=True)[1]
    return [params[ii] for ii in np.sort(msk)]


def filter_params_dup_hash(
    params: Sequence[dict], hashes: Sequence[str], **kwds
) -> Sequence[dict]:
    """Return params with psets whose hash is not in `hashes`.

    Use pset["_pset_hash"]  if present, else calculate hash on the fly.

    Parameters
    ----------
    params
    hashes
    kwds
        passed to :func:`pset_hash`
    """
    get_hash = lambda pset: pset.get("_pset_hash", pset_hash(pset, **kwds))
    return [pset for pset in params if get_hash(pset) not in hashes]


def stargrid(
    const: dict,
    vary: Sequence[Sequence[dict]],
    vary_labels: Sequence[str] = None,
    vary_label_col: str = "_vary",
    skip_dups=True,
) -> Sequence[dict]:
    """
    Helper to create a specific param sampling pattern.

    Vary params in a "star" pattern (and not a full pgrid) around constant
    values (middle of the "star").

    When doing that, duplicate psets can occur. By default try to filter them
    out (use filter_params_unique()) but ignore hash calculation errors and return
    non-reduced params in that case. If you want to fail at hash errors, use

    >>> filter_params_unique(stargrid(..., skip_dups=False), raise_error=True)

    Examples
    --------
    >>> import psweep as ps
    >>> const=dict(a=1, b=77, c=11)
    >>> a=ps.plist("a", [1,2,3,4])
    >>> b=ps.plist("b", [77,88,99])
    >>> c=ps.plist("c", [11,22,33,44])

    >>> ps.stargrid(const, vary=[a, b])
    [{'a': 1, 'b': 77, 'c': 11},
     {'a': 2, 'b': 77, 'c': 11},
     {'a': 3, 'b': 77, 'c': 11},
     {'a': 4, 'b': 77, 'c': 11},
     {'a': 1, 'b': 88, 'c': 11},
     {'a': 1, 'b': 99, 'c': 11}]

    >>> ps.stargrid(const, vary=[a, b], skip_dups=False)
    [{'a': 1, 'b': 77, 'c': 11},
     {'a': 2, 'b': 77, 'c': 11},
     {'a': 3, 'b': 77, 'c': 11},
     {'a': 4, 'b': 77, 'c': 11},
     {'a': 1, 'b': 77, 'c': 11},
     {'a': 1, 'b': 88, 'c': 11},
     {'a': 1, 'b': 99, 'c': 11}]

    >>> ps.stargrid(const, vary=[a, b], vary_labels=["a", "b"])
    [{'a': 1, 'b': 77, 'c': 11, '_vary': 'a'},
     {'a': 2, 'b': 77, 'c': 11, '_vary': 'a'},
     {'a': 3, 'b': 77, 'c': 11, '_vary': 'a'},
     {'a': 4, 'b': 77, 'c': 11, '_vary': 'a'},
     {'a': 1, 'b': 88, 'c': 11, '_vary': 'b'},
     {'a': 1, 'b': 99, 'c': 11, '_vary': 'b'}]

    >>> ps.stargrid(const, vary=[ps.itr2params(zip(a,c)),b], vary_labels=["a+c", "b"])
    [{'a': 1, 'b': 77, 'c': 11, '_vary': 'a+c'},
     {'a': 2, 'b': 77, 'c': 22, '_vary': 'a+c'},
     {'a': 3, 'b': 77, 'c': 33, '_vary': 'a+c'},
     {'a': 4, 'b': 77, 'c': 44, '_vary': 'a+c'},
     {'a': 1, 'b': 88, 'c': 11, '_vary': 'b'},
     {'a': 1, 'b': 99, 'c': 11, '_vary': 'b'}]

    >>> ps.stargrid(const, vary=[ps.pgrid([zip(a,c)]),b], vary_labels=["a+c", "b"])
    [{'a': 1, 'b': 77, 'c': 11, '_vary': 'a+c'},
     {'a': 2, 'b': 77, 'c': 22, '_vary': 'a+c'},
     {'a': 3, 'b': 77, 'c': 33, '_vary': 'a+c'},
     {'a': 4, 'b': 77, 'c': 44, '_vary': 'a+c'},
     {'a': 1, 'b': 88, 'c': 11, '_vary': 'b'},
     {'a': 1, 'b': 99, 'c': 11, '_vary': 'b'}]
    """
    params = []
    if vary_labels is not None:
        assert len(vary_labels) == len(
            vary
        ), f"{vary_labels=} and {vary=} must have same length"
    for ii, plist in enumerate(vary):
        for dct in plist:
            if vary_labels is not None:
                label = {vary_label_col: vary_labels[ii]}
                _dct = merge_dicts(dct, label)
            else:
                _dct = dct
            params.append(merge_dicts(const, _dct))

    if skip_dups:
        try:
            return filter_params_unique(
                params, raise_error=True, skip_special_cols=True
            )
        except PsweepHashError:
            return params
    else:
        return params


# -----------------------------------------------------------------------------
# run study
# -----------------------------------------------------------------------------

# tmpsave: That's cool, but when running in parallel, we loose the ability to
# store the whole state of the study calculated thus far. For that we would
# need an extra thread that periodically checks for or -- even better -- gets
# informed by workers about finished work and collects the so-far written temp
# results into a global df -- maybe useful for monitoring progress.


def worker_wrapper(
    pset: dict,
    worker: Callable,
    *,
    tmpsave: bool = False,
    verbose: bool | Sequence[str] = False,
    simulate: bool = False,
) -> dict:
    """
    Add special fields to `pset` which can be determined at call time.

    Call worker on exactly one pset. Return updated pset built from
    ``pset.update(worker(pset))``. Do verbose printing.
    """
    assert "_pset_id" in pset
    assert "_run_id" in pset
    assert "_calc_dir" in pset

    time_start = pd.Timestamp(time.time(), unit=PANDAS_TIME_UNIT)
    pset.update(_time_utc=time_start, _exec_host=platform.node())
    if verbose:
        df_row_print = pd.DataFrame([pset], index=[time_start])
        if isinstance(verbose, bool) and verbose:
            df_print(df_row_print, index=True)
        elif is_seq(verbose):
            df_print(df_row_print, index=True, cols=verbose)
        else:
            raise ValueError(f"Type of {verbose=} not understood.")
    t0 = time.time()
    if not simulate:
        pset.update(worker(pset))
    pset["_pset_runtime"] = time.time() - t0
    if tmpsave:
        fn = pj(
            pset["_calc_dir"],
            "tmpsave",
            pset["_run_id"],
            pset["_pset_id"] + ".pk",
        )
        pickle_write(fn, pset)
    return pset


def capture_logs_wrapper(
    pset: dict,
    worker: Callable,
    capture_logs: str,
    db_field: str = "_logs",
) -> dict:
    """Capture and redirect stdout and stderr produced in worker().

    Note the limitations mentioned in [1]:

        Note that the global side effect on sys.stdout means that this context
        manager is not suitable for use in library code and most threaded
        applications. It also has no effect on the output of subprocesses.
        However, it is still a useful approach for many utility scripts.

    So if users rely on playing with sys.stdout/stderr in worker(), then they
    sould not use this feature and take care of logging themselves.

    [1] https://docs.python.org/3/library/contextlib.html#contextlib.redirect_stdout
    """
    fn = f"{pset['_calc_dir']}/{pset['_pset_id']}/logs.txt"
    if capture_logs == "file":
        makedirs(os.path.dirname(fn))
        with open(fn, "w") as fd, redirect_stdout(fd), redirect_stderr(fd):
            return worker(pset)
    elif capture_logs in ["db", "db+file"]:
        with StringIO() as fd:
            with redirect_stdout(fd), redirect_stderr(fd):
                ret = worker(pset)
            txt = fd.getvalue()
        ret[db_field] = txt
        if capture_logs == "db+file":
            file_write(fn, txt)
        return ret
    else:
        raise ValueError(f"Illegal value {capture_logs=}")


def run(
    worker: Callable,
    params: Sequence[dict],
    df: pd.DataFrame = None,
    poolsize: int = None,
    dask_client=None,
    save: bool = True,
    tmpsave: bool = False,
    verbose: bool | Sequence[str] = False,
    calc_dir: str = "calc",
    simulate: bool = False,
    database_dir: str = None,
    database_basename: str = "database.pk",
    backup: bool = False,
    git: bool = False,
    skip_dups: bool = False,
    capture_logs: str = None,
) -> pd.DataFrame:
    """
    Call `worker` for each `pset` in `params`. Populate a DataFrame with rows
    from each call ``worker(pset)``.

    Parameters
    ----------
    worker
        must accept one parameter: `pset` (a dict ``{'a': 1, 'b': 'foo',
        ...}``), return either an update to `pset` or a new dict, result will
        be processes as ``pset.update(worker(pset))``
    params
        each dict is a pset ``{'a': 1, 'b': 'foo', ...}``
    df
        append rows to this DataFrame, if None then either create new one or
        read existing database file from disk if found
    poolsize
        * None : use serial execution
        * int : use multiprocessing.Pool (even for ``poolsize=1``)
    dask_client
        A dask client. Use this or ``poolsize``.
    save
        save final ``DataFrame`` to ``<database_dir>/<database_basename>``
        (pickle format only), default: "calc/database.pk", see also
        `database_dir`, `calc_dir` and `database_basename`
    tmpsave
        save the result dict from each ``pset.update(worker(pset))`` from each
        `pset` to ``<calc_dir>/tmpsave/<run_id>/<pset_id>.pk`` (pickle format
        only), the data is a dict, not a DataFrame row
    verbose
        * bool : print the current DataFrame row
        * sequence : list of DataFrame column names, print the row but only
          those columns
    calc_dir
        Dir where calculation artifacts can be saved if needed, such as dirs
        per pset ``<calc_dir>/<pset_id>``. Will be added to the database in
        ``_calc_dir`` field.
    simulate
        run everything in ``<calc_dir>.simulate``, don't call `worker`, i.e. save
        what the run would create, but without the results from `worker`,
        useful to check if `params` are correct before starting a production run
    database_dir
        Path for the database. Default is ``<calc_dir>``.
    database_basename
        ``<database_dir>/<database_basename>``, default: "database.pk"
    backup
        Make backup of ``<calc_dir>`` to ``<calc_dir>.bak_<timestamp>_run_id_<run_id>``
        where ``<run_id>`` is the latest ``_run_id`` present in ``df``
    git
        Use ``git`` to commit all files written and changed by the current run
        (``_run_id``). Make sure to create a ``.gitignore`` manually before if
        needed.
    skip_dups
        Skip psets whose hash is already present in `df`. Useful when repeating
        (parts of) a study.
    capture_logs
        {'db', 'file', 'db+file', None}
        Redirect stdout and stderr generated in ``worker()`` to database ('db')
        column ``_logs``, file ``<calc_dir>/<pset_id>/logs.txt``, or both. If
        ``None`` then do nothing (default). Useful for capturing per-pset log
        text, e.g. ``print()`` calls in `worker` will be captured.

    Returns
    -------
    df
        The database build from `params`.
    """

    # Don't in-place alter dicts in params we get as input.
    params = copy.deepcopy(params)

    database_dir = calc_dir if database_dir is None else database_dir

    git_enter(git)

    if simulate:
        calc_dir_sim = calc_dir + ".simulate"
        if os.path.exists(calc_dir_sim):
            shutil.rmtree(calc_dir_sim)
        makedirs(calc_dir_sim)
        old_db = pj(database_dir, database_basename)
        if os.path.exists(old_db):
            shutil.copy(old_db, pj(calc_dir_sim, database_basename))
        else:
            warnings.warn(
                f"simulate: {old_db} not found, will create new db in "
                f"{calc_dir_sim}"
            )
        database_fn = pj(calc_dir_sim, database_basename)
        calc_dir = calc_dir_sim
    else:
        database_fn = pj(database_dir, database_basename)

    if df is None:
        if os.path.exists(database_fn):
            df = df_read(database_fn)
        else:
            df = pd.DataFrame()

    if len(df) == 0:
        pset_seq_old = -1
        run_seq_old = -1
    else:
        pset_seq_old = df._pset_seq.values.max()
        run_seq_old = df._run_seq.values.max()

    if backup and len(df.index) > 0:
        stamp = df._time_utc.max().strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        dst = f"{calc_dir}.bak_{stamp}_run_id_{df._run_id.values[-1]}"
        assert not os.path.exists(dst), (
            "backup destination {dst} exists, seems like there has been no new "
            "data in {calc_dir} since the last backup".format(
                dst=dst, calc_dir=calc_dir
            )
        )
        shutil.copytree(calc_dir, dst)

    for pset in params:
        pset["_pset_hash"] = pset_hash(pset)

    if skip_dups and len(df) > 0:
        params = filter_params_dup_hash(params, df._pset_hash.values)

    run_id = get_uuid(existing=df._run_id.values if len(df) > 0 else [])
    pset_ids = get_many_uuids(
        len(params), existing=df._pset_id.values if len(df) > 0 else []
    )
    for ii, (pset, pset_id) in enumerate(zip(params, pset_ids)):
        pset["_pset_id"] = pset_id
        pset["_run_seq"] = run_seq_old + 1
        pset["_pset_seq"] = pset_seq_old + ii + 1
        pset["_run_id"] = run_id
        pset["_calc_dir"] = calc_dir

    if capture_logs is not None:
        worker = partial(
            capture_logs_wrapper, worker=worker, capture_logs=capture_logs
        )

    worker = partial(
        worker_wrapper,
        worker=worker,
        tmpsave=tmpsave,
        verbose=verbose,
        simulate=simulate,
    )

    if (poolsize is None) and (dask_client is None):
        results = list(map(worker, params))
    else:
        assert [poolsize, dask_client].count(
            None
        ) == 1, "Use either poolsize or dask_client."
        if dask_client is None:
            with mp.Pool(poolsize) as pool:
                results = pool.map(worker, params)
        else:
            futures = dask_client.map(worker, params)
            results = dask_client.gather(futures)

    df = pd.concat((df, pd.DataFrame(results)), sort=False, ignore_index=True)

    if save:
        df_write(database_fn, df)

    git_exit(git, df)

    return df


# -----------------------------------------------------------------------------
# (HPC cluster) batch runs using file templates
# -----------------------------------------------------------------------------


class Machine:
    def __init__(self, machine_dir: str, jobscript_name: str = "jobscript"):
        """
        Expected templates layout::

            templates/machines/<name>/info.yaml
            ^^^^^^^^^^^^^^^^^^^^^^^^^------------- machine_dir
            templates/machines/<name>/jobscript
                                      ^^^^^^^^^--- template.basename
        """
        self.name = os.path.basename(os.path.normpath(machine_dir))
        self.template = FileTemplate(
            pj(machine_dir, jobscript_name), target_suffix="_" + self.name
        )
        with open(pj(machine_dir, "info.yaml")) as fd:
            info = yaml.safe_load(fd)
        for key, val in info.items():
            assert key not in self.__dict__, f"cannot overwrite '{key}'"
            setattr(self, key, val)

    def __repr__(self):
        return f"{self.name}:{self.template}"


class FileTemplate:
    def __init__(self, filename, target_suffix=""):
        self.filename = filename
        self.basename = os.path.basename(filename)
        self.dirname = os.path.dirname(filename)
        self.targetname = f"{self.basename}{target_suffix}"

    def __repr__(self):
        return self.filename

    def fill(self, pset):
        try:
            return string.Template(file_read(self.filename)).substitute(pset)
        except:
            print(f"Failed to fill template: {self.filename}", file=sys.stderr)
            raise


def gather_calc_templates(calc_templ_dir):
    return [
        FileTemplate(pj(calc_templ_dir, basename))
        for basename in os.listdir(calc_templ_dir)
    ]


def gather_machines(machine_templ_dir):
    return [
        Machine(pj(machine_templ_dir, basename))
        for basename in os.listdir(machine_templ_dir)
    ]


# If we ever add a "simulate" kwd here: don't pass that thru to run() b/c
# there this prevents worker() from being executed, but that's what we always
# want here since it writes only input files. Instead, just set calc_dir =
# calc_dir_sim and copy the database as in run() and go. Don't copy the
# run_*.sh scripts b/c they are generated afresh anyway.
#
def prep_batch(
    params: Sequence[dict],
    *,
    calc_templ_dir: str = "templates/calc",
    machine_templ_dir: str = "templates/machines",
    git: bool = False,
    write_pset: bool = False,
    **kwds,
) -> pd.DataFrame:
    """
    Write files based on templates.

    Parameters
    ----------
    params
        See :func:`run`
    calc_templ_dir, machine_templ_dir
        Dir with templates.
    git
        Use git to commit local changes.
    write_pset
        Write the input `pset` to ``<calc_dir>/<pset_id>/pset.pk``.
    **kwds
        Passed to :func:`run`.

    Returns
    -------
    df
        The database build from `params`.
    """
    git_enter(git)

    calc_dir = kwds.get("calc_dir", "calc")

    calc_templates = gather_calc_templates(calc_templ_dir)
    machines = gather_machines(machine_templ_dir)
    templates = calc_templates + [m.template for m in machines]

    def worker(pset):
        for template in templates:
            file_write(
                pj(calc_dir, pset["_pset_id"], template.targetname),
                template.fill(pset),
            )
            if write_pset:
                pickle_write(pj(calc_dir, pset["_pset_id"], "pset.pk"), pset)
        return {}

    df = run(
        worker,
        params,
        git=False,
        **kwds,
    )

    msk_latest = df._run_seq == df._run_seq.values.max()
    msk_old = df._run_seq < df._run_seq.values.max()
    for machine in machines:
        txt = ""
        for pfx, msk in [("# ", msk_old), ("", msk_latest)]:
            if msk.any():
                txt += "\n"
            txt += "\n".join(
                f"{pfx}cd $here/{pset_id}; {machine.subcmd} {machine.template.targetname}  # run_seq={run_seq} pset_seq={pset_seq}"
                for pset_id, pset_seq, run_seq in zip(
                    df[msk]._pset_id.values,
                    df[msk]._pset_seq.values,
                    df[msk]._run_seq.values,
                )
            )
        file_write(
            f"{calc_dir}/run_{machine.name}.sh",
            f"#!/bin/sh\n\nhere=$(readlink -f $(dirname $0))\n{txt}\n",
        )

    git_exit(git, df)

    return df



import argparse
import json
import inspect

from . import psweep as ps


def check_calc_dir():
    func = ps.check_calc_dir
    desc = f"""
{inspect.getdoc(func)}

Uses {func.__name__}() in {inspect.getsourcefile(func)}.
"""
    parser = argparse.ArgumentParser(
        description=desc, formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument("calc_dir", metavar="CALC_DIR")
    args = parser.parse_args()
    calc_dir = args.calc_dir

    out = func(calc_dir, ps.df_read(f"{calc_dir}/database.pk"))
    _filt = lambda x: list(x) if isinstance(x, set) else x
    print(json.dumps(dict((kk, _filt(vv)) for kk, vv in out.items())))
